LFGTimer = LibStub("AceAddon-3.0"):NewAddon("LFGTimer", "AceEvent-3.0");
local addon = LFGTimer; -- upvalue the addon table
local L = LibStub("AceLocale-3.0"):GetLocale("LFGTimer", true);

local db;

function addon:OnInitialize()
	self:InitConfig(); -- setup config system
	db = self.db.profile; -- make sure this file has a valid db handle
	self:SetupFloatingFrame();
	self:RegisterEvent("LFG_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_SHOW", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_FAILED", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_SUCCEEDED", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_CHECK_SHOW", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_CHECK_HIDE", "UpdateFrameShown");
	self:RegisterEvent("LFG_BOOT_PROPOSAL_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("VOTE_KICK_REASON_NEEDED", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_UPDATE_RANDOM_INFO", "UpdateFrameShown");
	self:RegisterEvent("LFG_OPEN_FROM_GOSSIP", "UpdateFrameShown");
	self:RegisterEvent("GOSSIP_CLOSED", "UpdateFrameShown");
end

function addon:UpdateMainDB(database)
	db = database;
end


local _, addon = ...;

local L = LibStub("AceLocale-3.0"):GetLocale("LFGTimer");
local LSM = LibStub("LibSharedMedia-3.0");

if not addon then
	return;
end

LSM:Register("Background", "Blizzard Chat Bubble", "Interface\\Tooltips\\ChatBubble-Background");

local db; -- set after :InitConfig
local defaults = {
	profile = {
		ShowFrame = true,
		FrameType = "custom",
		Frame = {
			Tooltip = {
				Data = {
					Tank = true,
					Healer = true,
					DPS = true,
					Self = true,
					Party = true,
				},
				Backdrop = {
					Background = "Blizzard Chat Bubble",
					Border = "Blizzard Chat Bubble",
					BackgroundColor = {
						r = 0,
						g = 0,
						b = 0,
						a = 1,
					},
					BorderColor = {
						r = 1,
						g = 1,
						b = 1,
						a = 1,
					},
					Inset = 5,
					Tile = false,
					TileSize = 32,
					EdgeSize = 24,
				},
				Text = {
					Font = LSM:GetDefault("font"),
					Size = 12,
					Outline = " ",
					Color = {
						r = 1,
						g = 1,
						b = 1,
						a = 1,
					},
				},
			},
			ShowWhen = {
				Not = true,
				Acquire = true,
				Wait = true,
				Party = true,
				Found = true,
			},
			Position = {
				AnchorAtSelf = "CENTER",
				AnchoredTo = "UIParent",
				AnchorAtOther = "CENTER",
				AnchorXOffset = 0,
				AnchorYOffset = 0,
			},
		},
	},
};

function addon:InitConfig()
	self.db = LibStub("AceDB-3.0"):New("LFGTimerDB", defaults);
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged");
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged");
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged");
	
	db = self.db.profile;
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("LFGTimer", self.Options);
	
	local lib = LibStub("AceConfigDialog-3.0");
	self.OptionsFrames = {};
	
--	self.OptionsFrames.Main = lib:AddToBlizOptions("LFGTimer", "LFG Timer", nil, "General"); -- disable until more options are needed/added
--	self.OptionsFrames.Frame = lib:AddToBlizOptions("LFGTimer", L["Frame"], "LFG Timer", "Frame");
	
	self.OptionsFrames.Frame = lib:AddToBlizOptions("LFGTimer", "LFG Timer", nil, "Frame");
	self.Options.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db);
	self.OptionsFrames.Profiles = lib:AddToBlizOptions("LFGTimer", L["Profiles"], "LFG Timer", "Profiles");
end

function addon:OnProfileChanged(event, database)
	db = database.profile;
	self:UpdateFrameDB(db);
	self:UpdateMainDB(db);
end

addon.Options = {
	type = "group",
	order = 1,
	name = "LFG Timer",
	get = function(i)
		return db[ i[#i] ];
	end,
	set = function(i, v)
		db[ i[#i] ] = v;
	end,
	args = {
		General = {
			type = "group",
			order = 1,
			guiInline = true,
			name = L["General"],
			desc = L["General usage options"],
			args = {
				intro = {
					type = "description",
					name = L["LFGTimerAbout"],
					order = 1,
				},
			},
		},
		Frame = {
			type = "group",
			order = 2,
			guiInline = true,
			name = L["Frame"],
			desc = L["Floating frame options"],
			get = function(i)
				return db.Frame[ i[#i] ];
			end,
			set = function(i, v)
				db.Frame[ i[#i] ] = v;
				addon:UpdateFrameTextures();
			end,
			hidden = function()
				--[[
					hide when no frame is shown or the frame shown is not the custom frame
				--]]
				return not db.Frame.Show or db.Frame.Option ~= "custom";
			end,
			args = {
				Show = {
					type = "toggle",
					order = 1,
					name = L["Show"],
					desc = L["Show the floating frame while in an LFG queue"],
					set = function(i, v)
						db.Frame.Show = v;
						addon:UpdateFrameShown();
					end,
					width = "double",
					hidden = false,
				},
				ShowWhen = {
					type = "group",
					order = 2,
					guiInline = true,
					name = L["Show when"],
					desc = L["Control when the frame is shown"],
					set = function(i, v)
						db.Frame.ShowWhen[ i[#i] ] = v;
						addon:UpdateFrameShown();
					end,
					get = function(i)
						return db.Frame.ShowWhen[ i[#i] ];
					end,
					hidden = function()
						return not db.Frame.Show;
					end,
					args = {
						Not = {
							type = "toggle",
							order = 1,
							name = L["Not queued"],
							desc = L["Show the frame when you are not in an LFG queue"],
						},
						Acquire = {
							type = "toggle",
							order = 2,
							name = L["Acquiring data"],
							desc = L["Show the frame while acquring LFG queue data"],
						},
						Wait = {
							type = "toggle",
							order = 3,
							name = L["Waiting for a queue"],
							desc = L["Show the frame while waiting for the queue to pop"],
						},
						Party = {
							type = "toggle",
							order = 4,
							name = L["In a party"],
							desc = L["Show the frame while in an LFG party"],
						},
						Found = {
							type = "toggle",
							order = 5,
							name = L["Group found"],
							desc = L["Show the frame when a group is found"],
						},
					},
				},
				Option = {
					type = "select",
					order = 3,
					name = L["Frame to show"],
					desc = L["Choose which frame to show"],
					values = {
						["clone"] = L["Search status clone"],
						["custom"] = L["Custom frame"],
					},
					set = function(i, v)
						db.Frame.Option = v;
						addon:UpdateFrameShown();
					end,
					hidden = function()
						return not db.Frame.Show;
					end,
				},
				Data = {
					type = "group",
					order = 4,
					name = L["Information shown in the frame"],
					desc = L["Control what information is shown in the custom frame"],
					guiInline = true,
					get = function(i)
						return db.Frame[ i[#i] ];
					end,
					set = function(i, v)
						db.Frame[ i[#i] ] = v;
					end,
					args = {
						ShowTank = {
							type = "toggle",
							order = 1,
							name = L["Tank queue times"],
							desc = L["Show the average wait for people queued as a tank"],
						},
						ShowHealer = {
							type = "toggle",
							order = 2,
							name = L["Healer queue times"],
							desc = L["Show the average wait for people queued as a healer"],
						},
						ShowDPS = {
							type = "toggle",
							order = 3,
							name = L["DPS queue times"],
							desc = L["Show the average wait for people queued as a damage dealer"],
						},
						ShowSelf = {
							type = "toggle",
							order = 4,
							name = L["Your wait time"],
							desc = L["Show your estimated wait time"],
						},
						ShowParty = {
							type = "toggle",
							order = 5,
							name = L["Party members found"],
							desc = L["Show the number of party members found"],
						},
					},
				},
				Background = {
					type = "group",
					order = 5,
					guiInline = true,
					name = L["Frame background options"],
					get = function(i)
						return db.Frame.Background[ i[#i] ];
					end,
					set = function(i, v)
						db.Frame.Background[ i[#i] ] = v;
						addon:UpdateFrameTextures(); -- all of the options here affect the floating frame
					end,
					args = {
						Texture = {
							type = "select",
							dialogControl = "LSM30_Background",
							order = 2,
							name = L["Background texture"],
							desc = L["Texture to use for the frame's background"],
							values = AceGUIWidgetLSMlists.background,
						},
						Color = {
							type = "color",
							order = 1,
							name = L["Background color"],
							desc = L["Frame's background color"],
							hasAlpha = true,
							get = function(i)
								local t = db.Frame.Background.Color;
								return t.r, t.g, t.b, t.a;
							end,
							set = function(i, ...) -- red, green, blue, alpha
								local t = db.Frame.Background.Color;
								t.r, t.g, t.b, t.a = ...;
								addon:UpdateFrameTextures();
							end,
						},
						Tile = {
							type = "toggle",
							order = 5,
							name = L["Tile background"],
							desc = L["Tile the background texture"],
						},
						TileSize = {
							type = "range",
							order = 6,
							name = L["Background tile size"],
							desc = L["The size used to tile the background texture"],
							min = 16, 
							max = 256,
							step = 1,
							disabled = function()
								return not db.Frame.Background.Tile;
							end,
						},
					},
				},
				Border = {
					type = "group",
					order = 6,
					guiInline = true,
					name = L["Frame border options"],
					get = function(i)
						return db.Frame.Border[ i[#i] ];
					end,
					set = function(i, v)
						db.Frame.Border[ i[#i] ] = v;
						addon:UpdateFrameTextures();
					end,
					args = {
						Texture = {
							type = "select",
							dialogControl = "LSM30_Border",
							order = 2,
							name = L["Border texture"],
							desc = L["Texture to use for the frame's border"],
							values = AceGUIWidgetLSMlists.border,
						},
						Color = {
							type = "color",
							order = 1,
							name = L["Border color"],
							desc = L["Frame's border color"],
							hasAlpha = true,
							get = function(i)
								local t = db.Frame.Border.Color;
								return t.r, t.g, t.b, t.a;
							end,
							set = function(i, ...) -- red, green, blue, alpha
								local t = db.Frame.Border.Color;
								t.r, t.g, t.b, t.a = ...
								addon:UpdateFrameTextures();
							end,
						},
						EdgeSize = {
							type = "range",
							order = 3,
							name = L["Border thickness"],
							desc = L["The thickness of the border"],
							min = 1,
							max = 32,
							step = 1,
						},
					},
				},
				Text = {
					type = "group",
					order = 7,
					guiInline = true,
					name = L["Text options"],
					get = function(i)
						return db.Frame.Text[ i[#i] ];
					end,
					set = function(i, v)
						db.Frame.Text[ i[#i] ] = v;
						addon:UpdateFrameFonts();
					end,
					args = {
						Font = {
							type = "select",
							dialogControl = "LSM30_Font",
							order = 1,
							name = L["Font"],
							desc = L["Font used to display text"],
							values = AceGUIWidgetLSMlists.font,
						},
						Outline = {
							type = "select",
							order = 2,
							name = L["Font outline"],
							desc = L["Outline text in the frame"],
							values = {
								[" "] = L["None"],
								["OUTLINE"] = L["Outline"],
								["THICKOUTLINE"] = L["Thick outline"],
							},
						},
						Color = {
							type = "color",
							order = 3,
							name = L["Font color"],
							desc = L["Color of the font"],
							hasAlpha = true,
							get = function(i)
								local t = db.Frame.Text.Color;
								return t.r, t.g, t.b, t.a;
							end,
							set = function(i, ...) -- red, green, blue, alpha
								local t = db.Frame.Text.Color;
								t.r, t.g, t.b, t.a = ...;
							end
						},
						Size = {
							type = "range",
							order = 4,
							name = L["Font size"],
							desc = L["Set the font size of the text"],
							min = 4,
							max = 30,
							step = 1,
						},
					},
				},
			},
		},
	},
};


local _, addon = ...;

if not addon then
	return;
end

local LSM = LibStub("LibSharedMeda-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale("LFGTimer");

local db; -- set in :SetupFloatingFrame

function addon:UpateFrameDB(database)
	db = database;
end

function addon:SetupFloatingFrame()
	db = self.db.profile;
	self:UpdateFrameShown();
end

function addon:UpdateFrameShown()
	if self.CloneFrame then
		self.CloneFrame:Hide();
	end
	if self.CustomFrame then
		self.CustomFrame:Hide();
	end
	
	if self:ShouldShowFrame() then
		self:GetFrame():Show();
	end
end

function addon:GetFrame()
	if db.FrameType == "clone" then
		return self:SetupCloneFrame();
	elseif db.FrameType == "custom" then
		return self:SetupCustomFrame();
	end
end

function addon:SetupCloneFrame()
	-- Create clone frame
	local frame, f; -- Declare variables
	
	frame = CreateFrame("Frame", "LFGTimerSearchStatusClone", UIParent);
	
	frame:SetFrameStrata("TOOLTIP");
	frame:SetMovable(true);
	frame:SetClampedToScreen(true);
	
	frame:SetHeight(153);
	frame:SetWidth(350);
	
	frame:ClearAllPoints();
	frame:SetPoint("CENTER");
	
	frame.Backdrop = {
		bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
		edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {
			left = 5,
			right = 5,
			top = 5,
			bottom = 5,
		},
	};
	
	frame:SetBackdrop(frame.Backdrop);
	
	f = frame:CreateFontString("$parentTitle", "ARTWORK", "GameFontHighlight");
	f:SetText(ASSEMBLING_GROUP);
	f:SetPoint("TOP", 0, -15);
	frame.Title = f;
	
	f = frame:CreateFontString("$parentStatistic", "ARTWORK", "GameFontNormal");
	f:SetPoint("TOP", 0, -90);
	f:SetHeight(25);
	f:SetWidth(330);
	f:SetJustifyH("CENTER");
	f:SetJustifyV("MIDDLE");
	frame.Statistic = f;
	
	f = frame:CreateFontString("$parentElapsedWaitTime", "ARTWORK", "GameFontNormal");
	f:SetPoint("BOTTOM", 0, 38);
	f:SetJustifyH("CENTER");
	frame.WaitTime = f;
	
	f = frame:CreateFontString("$parentLookingFor", "ARTWORK", "GameFontHighlight");
	f:SetText(SEARCHING_FOR_GROUPS_NEEDS);
	f:SetPoint("BOTTOM", 0, 14);
	f:SetJustifyH("CENTER");
	frame.LookingFor = f;
	
	f = frame:CreateTexture("$parentRoleIcon1", "ARTWORK", "LFDSearchStatusRoleIconTemplate");
	f:ClearAllPoints();
	f:SetPoint("LEFT", frame.LookingFor, "RIGHT", 2, 0);
	f:SetHeight(25);
	f:SetWidth(25);
	frame.RoleIcon1 = f;
	
	f = frame:CreateTexture("$parentRoleIcon2", "ARTWORK", "LFDSearchStatusRoleIconTemplate");
	f:ClearAllPoints();
	f:SetPoint("LEFT", frame.RoleIcon1, "RIGHT", 2, 0);
	f:SetHeight(25);
	f:SetWidth(25);
	frame.RoleIcon2 = f;
	
	f = frame:CreateTexture("$parentRoleIcon3", "ARTWORK", "LFDSearchStatusRoleIconTemplate");
	f:ClearAllPoints();
	f:SetPoint("LEFT", frame.RoleIcon2, "RIGHT", 2, 0);
	f:SetHeight(25);
	f:SetWidth(25);
	frame.RoleIcon3 = f;
	
	frame.RoleIcons = {
		frame.RoleIcon1,
		frame.RoleIcon2,
		frame.RoleIcon3,
	};
	
	f = CreateFrame("Button", "$parentDragBar", frame);
	f:SetPoint("TOPLEFT", 10, -5);
	f:SetPoint("RIGHT", -10, 0);
	f:SetHeight(8);
	f:SetHighlightTexture([[Interface\FriendsFrame\UI-FriendsFrame-HighlightBar]]);
	f:SetScript("OnMouseDown", function(self, button)
		frame:StartMoving();
	end);
	f:SetScript("OnMouseUp", function(self, button)
		frame:StopMovingOrSizing();
		addon:UpdateFramePosition(frame);
	end);
	frame.Drag = f;
	
	f = CreateFrame("Frame", "$parentTank1", frame, "LFDSearchStatusPlayerTemplate");
	f:SetPoint("TOPLEFT", 25, -40);
	f.texture:SetTexCoord(GetTexCoordsForRole("TANK"));
	frame.Tank1 = f;
	
	f = CreateFrame("Frame", "$parentHealer1", frame, "LFDSearchStatusPlayerTemplate");
	f:SetPoint("LEFT", frame.Tank1, "RIGHT", 5, 0);
	f.texture:SetTexCoord(GetTexCoordsForRole("HEALER"));
	frame.Healer1 = f;
	
	f = CreateFrame("Frame", "$parentDamage1", frame, "LFDSearchStatusPlayerTemplate");
	f:SetPoint("LEFT", frame.Healer1, "RIGHT", 5, 0);
	f.texture:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
	frame.Damage1 = f;
	
	f = CreateFrame("Frame", "$parentDamage2", frame, "LFDSearchStatusPlayerTemplate");
	f:SetPoint("LEFT", frame.Damage1, "RIGHT", 5, 0);
	f.texture:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
	frame.Damage2 = f;
	
	f = CreateFrame("Frame", "$parentDamage3", frame, "LFDSearchStatusPlayerTemplate");
	f:SetPoint("LEFT", frame.Damage2, "RIGHT", 5, 0);
	f.texture:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
	frame.Damage3 = f;
	
	frame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	frame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	
	for k, v in pairs(addon.CloneFrameFuncs) do
		frame[k] = v;
	end
	
	frame:SetScript("OnShow", frame.OnShow);
	frame:SetScript("OnHide", frame.OnHide);
	
	self:SetAnchor(frame);
	
	self.CloneFrame = frame;
	frame:Hide();
	
	self.SetupCloneFrame = function(self)
		return self.CloneFrame;
	end -- Overwrite the setup funciton to just return the frame
	
	return frame;
end

function addon:SetupCustomFrame()
	-- Create custom data frame
	local frame, f;

	frame = CreateFrame("Frame", "LFGTimerCustomDataFrame", UIParent);

	frame:SetFrameStrata("MEDIUM");
	frame:SetMovable(true);
	frame:SetUserPlaced(true);
	frame:SetClampedToScreen(true);

	frame:ClearAllPoints();
	frame:SetPoint("CENTER");
	frame:SetHeight(150);
	frame:SetWidth(150);

	self:SetAnchor(frame);

	f = CreateFrame("Button", "$parentDragBar", frame);
	f:SetPoint("TOP", 0, -5);
	f:SetPoint("LEFT", 10, 0);
	f:SetPoint("RIGHT", -10, 0);
	f:SetHeight(8);
	f:SetHighlightTexture([[Interface\FriendsFrame\UI-FriendsFrame-HighlightBar]]);
	f:SetScript("OnMouseDown", function(self, button)
		frame:StartMoving();
	end);
	f:SetScript("OnMouseUp", function(self, button)
		frame:StopMovingOrSizing();
		addon:UpdateFramePosition(frame);
	end);
	frame.Drag = f;

	f = CreateFrame("Frame", "$parentContainer", frame);
	f:SetPoint("TOP", frame.Drag, "BOTTOM", 0, -5);
	f:SetPoint("LEFT", 5, 0);
	f:SetPoint("RIGHT", -5, 0);
	f:SetPoint("BOTTOM", 0, 5);
	frame.Container = f;

	frame.Lines = {};
	frame.TextFrames = {};
	frame.NumTextFrames = 0;
	frame.Backdrop = {
		insets = {},
	};

	for k, v in pairs(addon.CustomFrameFuncs) do
		frame[k] = v;
	end

	frame:SetScript("OnShow", frame.OnShow);
	frame:SetScript("OnHide", frame.OnHide);

	frame:Hide();

	frame:UpdateFrameArt();

	self.CustomFrame = frame;
	
	self.SetupCustomFrame = function(self)
		return self.CustomFrame;
	end -- Overwrite the setup funciton to just return the frame
	
	return frame;
end

addon.CloneFrameFuncs = {
	OnShow = function(frame)
		frame:SetScript("OnUpdate", frame.Update);
	end,
	OnHide = function(frame)
		frame:SetScript("OnUpdate", nil);
	end,
	Update = function(frame)
		local lfgData, lfgMode = addon:UpdateQueueData();
	
		frame:SetHeight(153);
		frame.WaitTime:Hide();
		
		addon.CloneFrame.UpdateRoles(frame);
		
		if (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then
			frame.Statistic:SetText(L["You are not in an LFG queue"]);
		elseif lfgMode[1] == "proposal" then
			frame.Statistic:SetText(L["Found a group"]);
			if lfgMode[2] == "accepted" then
				frame.Statistic:SetText(L["Found a group"] .. "\n" .. L["Waiting on group members"]);
				frame:SetHeight(170);
			end
		elseif (lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then
			frame.Statistic:SetText(L["Acquiring LFG queue data..."]);
		elseif lfgMode[1] == "lfgparty" then
			frame.Statistic:SetText(L["You are in an LFG party"]);
		else
			local elapsedTime = GetTime() - lfgData[13];
			frame.Statistic:SetFormattedText(LFG_STATISTIC_AVERAGE_WAIT, lfgData[12] == -1 and TIME_UNKNOWN or SecondsToTime(lfgData[12], false, false, 1));
			frame.WaitTime:SetFormattedText(TIME_IN_QUEUE, (elapsedTime >= 60) and SecondsToTime(elapsedTime) or LESS_THAN_ONE_MINUTE);
			frame.WaitTime:Show();
			frame:SetHeight(158);
		end
	end,
	UpdateRoleIcon = function(icon, need)
		if need then
			SetDesaturation(icon.texture, true);
			icon.cover:Show();
		else
			SetDesaturation(icon.texture, false);
			icon.cover:Hide();
		end
	end,
	UpdateRoles = function(frame)
		local UpdateRoleIcon = addon.CloneFrame.UpdateRoleIcon;
		local lfgData, lfgMode = addon.LFGData, addon.LFGMode;
		local leader, tank, healer, damage = GetLFGRoles();
		local icon, i;
		
		UpdateRoleIcon(frame.Tank1, lfgData[3]);
		UpdateRoleIcon(frame.Healer1, lfgData[4]);
		for i = 3, 1, -1 do
			icon = frame["Damage" .. i];
			UpdateRoleIcon(icon, i > lfgData[5]);
		end
		
		i = 1;
		if tank then
			icon = frame["RoleIcon" .. i];
			icon:SetTexCoord(GetTexCoordsForRole("TANK"));
			icon:Show();
			i = i + 1;
		end
		if healer then
			icon = frame["RoleIcon" .. i];
			icon:SetTexCoord(GetTexCoordsForRole("HEALER"));
			icon:Show();
			i = i + 1;
		end
		if damage then
			icon = frame["RoleIcon" .. i];
			icon:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
			icon:Show();
			i = i + 1;
		end
		local extraWidth = 27 * (i - 1);
		while i <= 3 do
			icon = frame["RoleIcon" .. i];
			icon:Hide();
			i = i + 1;
		end
		frame.LookingFor:SetPoint("BOTTOM", -extraWidth / 2, 14);
	end,
};

addon.CustomFrameFuncs = {
	OnShow = function(frame)
		frame:SetScript("OnUpdate", frame.UpdateData);
	end,
	OnHide = function(frame)
		frame:SetScript("OnUpdate", nil);
	end,
	UpdateFrameArt = function(frame)
		local inset = db.Frame.Tooltip.Backdrop.Inset;
		local backdrop = frame.Backdrop;
		
		backdrop.bgFile = LSM:Fetch("background", db.Frame.Tooltip.Backdrop.Background);
		backdrop.edgeFile = LSM:Fetch("border", db.Frame.Tooltip.Backdrop.Border);
		backdrop.tile = db.Frame.Tooltip.Backdrop.Tile;
		backdrop.tileSize = db.Frame.Tooltip.Backdrop.TileSize;
		backdrop.edgeSize = db.Frame.Tooltip.Backdrop.EdgeSize;
		
		backdrop.insets.left = inset;
		backdrop.insets.right = inset;
		backdrop.insets.top = inset;
		backdrop.insets.bottom = inset;
		
		local bgColor, borderColor = db.Frame.Tooltip.Backdrop.BackgroundColor, db.Frame.Tooltip.Backdrop.BorderColor;
		
		frame:SetBackdrop(backdrop);
		frame:SetBackdropColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);
		frame:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, borderColor.a);
		
		frame.Drag:SetPoint("TOP", 0, -inset);
		frame.Drag:SetPoint("LEFT", inset, 0);
		frame.Drag:SetPoint("RIGHT", -inset, 0);
		
		frame.Container:ClearAllPoints();
		frame.Container:SetPoint("TOP", frame.Drag, "BOTTOM", 0, -5);
		frame.Container:SetPoint("BOTTOM", 0, inset);
		frame.Container:SetPoint("LEFT", inset, 0);
		frame.Container:SetPoint("RIGHT", -inset, 0);
		
		frame:UpdateSize();
	end,
	ClearLines = function(frame)
		for k, v in pairs(frame.Lines) do
			frame.Lines[k] = nil;
			v:Hide();
			v.Text:SetText("Empty");
			frame.TextFrames[v] = true;
		end
	end,
	GetTextFrame = function(frame)
		local f = next(frame.TextFrames);
		if not f then
			frame.NumTextFrames = frame.NumTextFrames + 1;
			f = CreateFrame("Frame", "LFGTimerTextFrame" .. frame.NumTextFrames, frame.Container, "LFGTimerTextFrameTemplate");
		else
			frame.TextFrames[f] = nil;
		end
		
		local font, size, flags, color = (LSM:Fetch("font", db.Frame.Tooltip.Text.Font)), db.Frame.Tooltip.Text.Size, db.Frame.Tooltip.Text.Outline, db.Frame.Tooltip.Text.Color;
		
		f.Text:SetText("Initialized");
		f.Text:Show();
		f.Text:SetFont(font, size, flags);
		f.Text:SetTextColor(color.r, color.g, color.b, color.a);
		return f;
	end,
	AddLine = function(frame, text)
		local f = frame:GetTextFrame();
		f.Text:SetText(text);
		table.insert(frame.Lines, f);
	end,
	UpdateSize = function(frame)
		local inset = db.Frame.Tooltip.Backdrop.Inset;
		local last = nil;
		local maxWidth, height = 0, (2 * inset) + frame.Drag:GetHeight() + 10;
		
		for i, v in ipairs(frame.Lines) do
			v:Show();
			v.Text:Show();
			v:ClearAllPoints();
			if not last then
				v:SetPoint("TOP", frame.Container, "TOP", 0, 0);
			else
				v:SetPoint("TOP", last, "BOTTOM", 0, 0);
			end
			v:SetPoint("LEFT", frame.Container, "LEFT", inset, 0);
			v:SetWidth(v.Text:GetStringWidth() + (2 * inset));
			v:SetHeight(v.Text:GetStringHeight());
			height = height + v:GetHeight();
			
			if v:GetWidth() > maxWidth then
				maxWidth = v:GetWidth();
			end
			last = v;
		end
		
		frame:SetWidth(maxWidth + (2 * inset));
		frame:SetHeight(height);
	end,
	UpdateData = function(frame)
		local lfgData, lfgMode = addon:UpdateQueueData();
		frame:ClearLines();
		
		if (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then
			frame:AddLine(L["You are not in an LFG queue"]);
		elseif lfgMode[1] == "proposal" then
			frame:AddLine(L["Found a group"]);
			if lfgMode[2] == "accepted" then
				frame:AddLine(L["Waiting on group members"]);
			end
		elseif (lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then
			frame:AddLine(L["Acquiring LFG queue data..."]);
			if lfgData[13] then
				self:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData[13])));
			end
		elseif lfgMode[1] == "lfgparty" then
			frame:AddLine(L["You are in an LFG party"]);
		elseif lfgMode[1] == "queued" and lfgData[1] then
			if db.Frame.Tooltip.Data.Tank then
				frame:AddLine((L["Tank queue times"] .. ": %s"):format(lfgData[9] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[9])));
			end
			if db.Frame.Tooltip.Data.Healer then
				frame:AddLine((L["Healer queue times"] .. ": %s"):format(lfgData[10] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[10])));
			end
			if db.Frame.Tooltip.Data.DPS then
				frame:AddLine((L["DPS queue times"] .. ": %s"):format(lfgData[11] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[11])));
			end
			if db.Frame.Tooltip.Data.Self then
				frame:AddLine((L["Your queue time"] .. ": %s"):format(lfgData[12] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[12])));
			end
			frame:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData[13])));
			if db.Frame.Tooltip.Data.Party then
				frame:AddLine((L["Party members found"] ..": %d/5"):format(5 - lfgData[3] - lfgData[4] - lfgData[5]));
			end
		else
			self:AddLine(L["Found an odd state."]);
			self:AddLine(L["Waiting on updated data from the server."]);
		end
		
		frame:UpdateSize();
	end,
};








function addon:CreateCloneFrame()
	if self.Clone then
		self.Clone:Show();
		return;
	end
	self.Clone = CreateFrame("Frame", "LFGTimerCloneFrame", UIParent);
	local frame = self.Clone;
	local name = frame:GetName();
	frame:SetWidth(350);
	frame:SetHeight(153);
	frame:SetFrameStrata("TOOLTIP");
	frame:SetMovable(true);
	frame:SetClampedToScreen(true);
	frame:SetUserPlaced(true);
	self:SetAnchor(self.Clone); -- put the frame where it should be
	frame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true,
		edgeSize = 16,
		tileSize = 16,
		insets = {
			left = 5,
			right = 5,
			top = 5,
			bottom = 5,
		},
	});
	
	frame.Drag = CreateFrame("Button", frame:GetName() .. "DragButton", frame);
	frame.Drag:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -5);
	frame.Drag:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -5);
	frame.Drag:SetHeight(8);
	frame.Drag:SetHighlightTexture("Interface\\FriendsFrame\\UI-FriendsFrame-HighlightBar");
	frame.Drag:SetScript("OnMouseDown", function()
		frame:StartMoving()
	end);
	frame.Drag:SetScript("OnMouseUp", function()
		frame:StopMovingOrSizing();
		addon:UpdatePosition(addon.Clone);
	end);
	
	frame.Title = frame:CreateFontString(name .. "Title", "ARTWORK", "GameFontHighlight");
	frame.Title:SetText(ASSEMBLING_GROUP);
	frame.Title:SetPoint("TOP", frame, "TOP", 0, -15);
	
	frame.Statistic = frame:CreateFontString(name .. "Statistic", "ARTWORK", "GameFontNormal");
	frame.Statistic:SetHeight(25);
	frame.Statistic:SetWidth(330);
	frame.Statistic:SetJustifyH("CENTER");
	frame.Statistic:SetJustifyV("MIDDLE");
	frame.Statistic:SetPoint("TOP", frame, "TOP", 0, -90);
	
	frame.WaitTime = frame:CreateFontString(name .. "WaitTime", "ARTWORK", "GameFontNormal");
	frame.WaitTime:ClearAllPoints();
	frame.WaitTime:SetPoint("BOTTOM", frame, "BOTTOM", 0, 38);
	frame.WaitTime:Hide();
	
	frame.LookingFor = frame:CreateFontString(name .. "LookingFor", "ARTWORK", "GameFontHighlight");
	frame.LookingFor:SetText(SEARCHING_FOR_GROUPS_NEEDS);
	frame.LookingFor:SetJustifyH("CENTER");
	frame.LookingFor:SetPoint("BOTTOM", frame, "BOTTOM", 0, 14);
	
	local last;
	
	for i = 1, 3 do
		local tex = frame:CreateTexture(name .. "RoleIcon" .. i, "ARTWORK", "LFDSearchStatusRoleIconTemplate");
		tex:ClearAllPoints();
		if i == 1 then -- first role icon
			tex:SetPoint("LEFT", frame.LookingFor, "RIGHT", 2, 0);
		else
			tex:SetPoint("LEFT", last, "RIGHT", 2, 0);
		end
		last = tex; --]]
		frame["RoleIcon" .. i] = tex;
	end
	
	frame.Tank1 = CreateFrame("Frame", name .. "Tank1", frame, "LFDSearchStatusPlayerTemplate");
	frame.Tank1:SetPoint("TOPLEFT", frame, "TOPLEFT", 25, -40);
	frame.Tank1.texture:SetTexCoord(GetTexCoordsForRole("TANK"));
	
	frame.Healer1 = CreateFrame("Frame", name .. "Healer1", frame, "LFDSearchStatusPlayerTemplate");
	frame.Healer1:SetPoint("LEFT", frame.Tank1, "RIGHT", 5, 0);
	frame.Healer1.texture:SetTexCoord(GetTexCoordsForRole("HEALER"));
	
	last = frame.Healer1;
	
	for i = 1, 3 do
		local f = CreateFrame("Frame", name .. "Damage" .. i, frame, "LFDSearchStatusPlayerTemplate");
		f:SetPoint("LEFT", last, "RIGHT", 5, 0);
		f.texture:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
		frame["Damage" .. i] = f;
		last = f;
	end
	
	last = nil;
	
	self:CloneUpdateRoles();
	
	frame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	frame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	frame:SetScript("OnUpdate", self.StringHooks["CloneUpdate"]);
end

function addon:CloneUpdate()
	local frame = self.Clone;
	local lfgData = Recycler:Acquire(GetLFGQueueStats());
	local lfgMode = Recycler:Acquire(GetLFGMode());
--[[	local lfgData = {GetLFGQueueStats()};
	local lfgMode = {GetLFGMode()}; --]]
	frame:SetHeight(153); -- reset frame height
	
	self:CloneUpdateRoles();
	
	if (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then
		frame.Statistic:SetText(L["You are not in an LFG queue"]);
		frame.WaitTime:Hide();
	elseif lfgMode[1] == "proposal" then
		frame.Statistic:SetText(L["Found a group"]);
		if lfgMode[2] == "accepted" then
			frame.Statistic:SetText(L["Found a group"] .. "\n" .. L["Waiting on group members"]);
			frame:SetHeight(170); -- set for 3 lines of text
		end
	elseif(lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then -- queued, waiting on data
		if lfgData[13] then
			local t = GetTime() - lfgData[13];
			frame.WaitTime:SetFormattedText(TIME_IN_QUEUE, (t >= 60) and SecondsToTime(t) or LESS_THAN_ONE_MINUTE);
			frame.WaitTime:Show();
			frame:SetHeight(158);
		end
		frame.Statistic:SetText(L["Acquiring LFG queue data..."]);
	elseif lfgMode[1] == "lfgparty" then -- in an instance
		frame.Statistic:SetText(L["You are in an LFG party"]);
	else
		local elapsedTime = GetTime() - lfgData[13];
		frame.Statistic:SetFormattedText(LFG_STATISTIC_AVERAGE_WAIT, lfgData[12] == -1 and TIME_UNKNOWN or SecondsToTime(lfgData[12], false, false, 1));
		frame.WaitTime:SetFormattedText(TIME_IN_QUEUE, (elapsedTime >= 60) and SecondsToTime(elapsedTime) or LESS_THAN_ONE_MINUTE);
		frame.WaitTime:Show();
		frame:SetHeight(158);
--		frame.Statistic:SetFormattedText(LFG_STATISTIC_AVERAGE_WAIT .. "\n" .. L["Time in queue: %s"], lfgData[12] == -1 and TIME_UNKNOWN or SecondsToTime(lfgData[12], false, false, 1), SecondsToTime(GetTime() - db.LFGStartTime, false, false, 1));
	end
	
	if lfgData[3] == nil then
		lfgData[3] = 1;
	end
	if lfgData[4] == nil then
		lfgData[4] = 1;
	end
	if lfgData[5] == nil then
		lfgData[5] = 3;
	end
	
	self:CloneUpdateFound({
		tank = (1 - lfgData[3]),
		healer = (1 - lfgData[4]),
		dps = (3 - lfgData[5]),
	});
	Recycler:Release(lfgData);
	Recycler:Release(lfgMode);
end

function addon:CloneUpdateFound(t)
--[[
	t: # of group members found
	t.tank: # of tanks
	t.healer: # of healers
	t.dps: # of damage dealers
--]]
	local frame = self.Clone;
	if t.tank > 0 then -- have a tank
		SetDesaturation(frame.Tank1.texture, false);
		frame.Tank1.cover:Hide();
	else
		SetDesaturation(frame.Tank1.texture, true);
		frame.Tank1.cover:Show();
	end
	if t.healer > 0 then -- have a tank
		SetDesaturation(frame.Healer1.texture, false);
		frame.Healer1.cover:Hide();
	else
		SetDesaturation(frame.Healer1.texture, true);
		frame.Healer1.cover:Show();
	end
	for i = 1, 3 do
		local f = frame["Damage" .. i];
		if t.dps >= i then
			SetDesaturation(f.texture, false);
			f.cover:Hide();
		else
			SetDesaturation(f.texture, true);
			f.cover:Show();
		end
	end
end

function addon:CloneUpdateRoles()
	local frame = self.Clone;
	local roles = Recycler:Acquire(GetLFGRoles());
--	local roles = {GetLFGRoles()};
	local i = 1;
	local icon;
	if roles[2] then
		icon = frame["RoleIcon" .. i];
		icon:SetTexCoord(GetTexCoordsForRole("TANK"));
		icon:Show();
		i = i + 1;
	end
	if roles[3] then
		icon = frame["RoleIcon" .. i];
		icon:SetTexCoord(GetTexCoordsForRole("HEALER"));
		icon:Show();
		i = i + 1;
	end
	
	if roles[4] then
		icon = frame["RoleIcon" .. i];
		icon:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
		icon:Show();
		i = i + 1;
	end
	local eW = 27 * (i - 1); -- extra width
	while i <= 3 do
		icon = frame["RoleIcon" .. i];
		icon:Hide();
		i = i + 1;
	end
	frame.LookingFor:SetPoint("BOTTOM", -eW / 2, 14);
	Recycler:Release(roles);
end

function addon:CreateCustomFrame()
	if addon.Frame then -- already created the frame
		addon.Frame:Show(); -- so show it
		return
	end
	self.Frame = CreateFrame("Frame", "LFGTimerCustomFrame", UIParent);
	self.Frame.Lines = {};
	self.TextFrames = {};
	local frame = addon.Frame;
	
	frame:SetMovable(true);
	frame:SetClampedToScreen(true);
	frame:SetScript("OnUpdate", function()
		addon:UpdateFrameData();
	end);
	frame:SetUserPlaced(true);
	frame:SetWidth(150);
	frame:SetHeight(150);
	self:SetAnchor(self.Frame); -- put the frame where it should be
	
	frame.Drag = CreateFrame("Button", frame:GetName() .. "DragButton", frame);
	frame.Drag:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -5);
	frame.Drag:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -5);
	frame.Drag:SetHeight(8);
	frame.Drag:SetHighlightTexture("Interface\\FriendsFrame\\UI-FriendsFrame-HighlightBar");
	frame.Drag:SetScript("OnMouseDown", function()
		frame:StartMoving()
	end);
	frame.Drag:SetScript("OnMouseUp", function()
		frame:StopMovingOrSizing();
		addon:UpdatePosition(addon.Frame);
	end);
	
	local inset = addon:CalcInset();
	
	frame.Container = CreateFrame("Frame", frame:GetName() .. "TextContainer", frame);
	
	frame.Container:SetPoint("TOP", frame.Drag, "BOTTOM", 0, -5);
	frame.Container:SetPoint("BOTTOM", frame, "BOTTOM", 0, inset);
	frame.Container:SetPoint("LEFT", frame, "LEFT", inset, 0);
	frame.Container:SetPoint("RIGHT", frame, "RIGHT", -inset, 0);
	
	self:UpdateFrameData();
	self:UpdateFrameTextures();
	self:UpdateFrameFonts();
end

function addon:ShowFrame() -- returns if the frame should be shown
	local lfgData = Recycler:Acquire(GetLFGQueueStats());
	local lfgMode = Recycler:Acquire(GetLFGMode());
	local ret = false;
--[[	local lfgData = {GetLFGQueueStats()}; -- possibly cache these values and update on an event (LFG_UPDATE maybe)
	local lfgMode = {GetLFGMode()}; --]]
	if not db.Frame.Show then
		ret = false;
	elseif (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then -- not queued
		ret = db.Frame.ShowWhen.Not;
	elseif lfgMode[1] == "proposal" then -- queue popped, waiting to join
		ret = db.Frame.ShowWhen.Found;
	elseif (lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then -- queued, waiting on data
		ret = db.Frame.ShowWhen.Acquire;
	elseif lfgMode[1] == "lfgparty" then -- in an instance
		ret = db.Frame.ShowWhen.Party;
	elseif lfgMode[1] == "queued" and lfgData[1] then -- queued, have data
		ret = db.Frame.ShowWhen.Wait;
	else
		ret = false; -- wtf is going on with LFG, hide the frame
	end
	Recycler:Release(lfgData);
	Recycler:Release(lfgMode);
	return ret;
end

function addon:UpdateFrameShown()
	if self.Clone then
		self.Clone:Hide();
	end
	if self.Frame then
		self.Frame:Hide();
	end
	if self:ShowFrame() then
		if db.Frame.Option == "clone" then
			if not self.Clone then
				addon:CreateCloneFrame();
			end
			self.Clone:Show();
		elseif db.Frame.Option == "custom" then
			if not self.Frame then
				addon:CreateCustomFrame();
			end
			self.Frame:Show();
		end
	end
end

function addon:CalcInset()
	if db.Frame.Border.Texture == "None" then
		return 0;
	elseif db.Frame.Border.Texture == "Blizzard Chat Bubble" then -- extra wide border texture
		return db.Frame.Border.EdgeSize;
	else
		return math.floor(db.Frame.Border.EdgeSize / 2);
	end
end

function addon:ClearLines() -- hide all lines of text and empty the list of lines on the frame
	if not self.Frame then -- no custom frame. why was this called?
		return;
	end
	Recycler:Release(self.Frame.Lines);
	self.Frame.Lines = Recycler:Acquire();
	for i, v in ipairs(self.TextFrames) do
		v:Hide();
		v.Text:SetText("Empty");
	end
end

function addon:GetTextFrame(i)
	if not self.TextFrames[i] then
		local font = lsm:Fetch("font", db.Frame.Text.Font);
		local size = db.Frame.Text.Size;
		local flags = db.Frame.Text.Outline;
		local color = db.Frame.Text.Color;
		local frame = CreateFrame("Frame", "LFGTimerTextFrame" .. i, self.Frame);
		frame.Text = frame:CreateFontString(frame:GetName() .. "Text", nil, "GameFontNormalSmall"); -- starter font object
		frame.Text:SetJustifyH("LEFT");
		frame.Text:SetJustifyV("TOP");
		frame.Text:SetText("Initialized");
		frame.Text:SetFont(font, size, flags);
		frame.Text:SetTextColor(color.r, color.g, color.b, color.a);
		frame.Text:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5); -- bind the text to the top left corner of the frame. will be "broken" for right-to-left locales
		frame.Text:Show(); -- make sure the text is shown
		self.TextFrames[i] = frame;
	end
	return self.TextFrames[i];
end

function addon:AddLine(text) -- add a line of text to the custom frame
	if not self.Frame then -- why are we adding lines to a frame that does not exist?
		return;
	end
	local frame = self:GetTextFrame(#self.Frame.Lines + 1);
	frame.Text:SetText(text);
	table.insert(self.Frame.Lines, frame);
end

function addon:UpdateSize()
	local inset = self:CalcInset();
	local last = nil;
	local maxWidth, height = 0, 25;
	for i, v in pairs(self.Frame.Lines) do
		v:ClearAllPoints();
		if not last then -- first line
			v:SetPoint("TOP", self.Frame.Container, "TOP", 0, inset / 3);
		else
			v:SetPoint("TOP", last, "BOTTOM", 0, 0);
		end
		v:SetPoint("LEFT", self.Frame.Container, "LEFT", (-2 * inset / 3), 0);
		v:SetWidth(v.Text:GetWidth() + (inset / 3));
		v:SetHeight(v.Text:GetHeight());
		height = height + v:GetHeight();
		if v:GetWidth() > maxWidth then
			maxWidth = v:GetWidth();
		end
		v:Show();
		last = v;
	end
	self.Frame:SetWidth(maxWidth + 18); -- add offset to allow for extra room in the frame
	self.Frame:SetHeight(height);
end

function addon:UpdatePosition(frame)
	local anchor = db.Frame.Position;
	local x, y = 0, 0;
	local c1, c2 = {UIParent:GetCenter()}, {frame:GetCenter()};
	x = c2[1] - c1[1];
	y = c2[2] - c1[2];
	anchor.AnchorAtSelf = "CENTER";
	anchor.AnchoredTo = "UIParent";
	anchor.AnchorAtOther = "CENTER";
	anchor.AnchorXOffset = x;
	anchor.AnchorYOffset = y;
	self:SetAnchor(frame);
end

function addon:SetAnchor(frame)
	local anchor = db.Frame.Position;
	frame:ClearAllPoints();
	frame:SetPoint(anchor.AnchorAtSelf, anchor.AnchoredTo, anchor.AnchorAtOther, anchor.AnchorXOffset, anchor.AnchorYOffset);
end

function addon:UpdateFrameData()
	self:ClearLines();
	local lfgData = Recycler:Acquire(GetLFGQueueStats());
	local lfgMode = Recycler:Acquire(GetLFGMode());
	if (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then -- not queued
		self:AddLine(L["You are not in an LFG queue"]);
	elseif lfgMode[1] == "proposal" then -- queue popped, waiting to join
		self:AddLine(L["Found a group"]);
		if lfgMode[2] == "accepted" then
			self:AddLine(L["Waiting on group members"]);
		end
	elseif (lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then -- queued, waiting on data
		self:AddLine(L["Acquiring LFG queue data..."]);
		if lfgData[13] then
			self:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData[13])));
		end
	elseif lfgMode[1] == "lfgparty" then -- in an instance
		self:AddLine(L["You are in an LFG party"]);
	elseif lfgMode[1] == "queued" and lfgData[1] then -- queued, have data
		if db.Frame.ShowTank then
			self:AddLine((L["%s queue times: %s"]):format(L["Tank"], lfgData[9] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[9])));
		end
		if db.Frame.ShowHealer then
			self:AddLine((L["%s queue times: %s"]):format(L["Healer"], lfgData[10] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[10])));
		end
		if db.Frame.ShowDPS then
			self:AddLine((L["%s queue times: %s"]):format(L["DPS"], lfgData[11] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[11])));
		end
		if db.Frame.ShowSelf then
			self:AddLine((L["Estimated queue time: %s"]):format(lfgData[12] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[12])));
		end
		self:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData[13])));
		if db.Frame.ShowParty then
			self:AddLine((L["Party members found: %d/%d"]):format(TOTAL_LFG_PARTY_SIZE - lfgData[3] - lfgData[4] - lfgData[5], TOTAL_LFG_PARTY_SIZE));
		end
	else
		self:AddLine(L["Found an odd state, please submit the data that the addon prints to the chat window"]);
		self:TableDump({GetLFGMode = lfgMode, GetLFGQueueStats = lfgData}); -- ask people to submit this, i'm not able to produce it and i believe that some other states should be handled
	end
	self:UpdateSize();
	Recycler:Release(lfgData);
	Recycler:Release(lfgMode);
end

function addon:UpdateFrameTextures()
	local inset = self:CalcInset();
	local backdrop = {
		bgFile = lsm:Fetch("background", db.Frame.Background.Texture);
		edgeFile = lsm:Fetch("border", db.Frame.Border.Texture);
		tile = db.Frame.Background.Tile,
		tileSize = db.Frame.Background.TileSize,
		edgeSize = db.Frame.Border.EdgeSize,
		insets = {
			left = inset,
			right = inset,
			top = inset,
			bottom = inset,
		},
	};
	local bgColor = db.Frame.Background.Color;
	local borderColor = db.Frame.Border.Color;
	
	self.Frame:SetBackdrop(backdrop);
	self.Frame:SetBackdropColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);
	self.Frame:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, borderColor.a);
	
	self.Frame.Container:ClearAllPoints();
	self.Frame.Container:SetPoint("TOP", self.Frame.Drag, "BOTTOM", 0, -5);
	self.Frame.Container:SetPoint("BOTTOM", self.Frame, "BOTTOM", 0, inset);
	self.Frame.Container:SetPoint("LEFT", self.Frame, "LEFT", inset, 0);
	self.Frame.Container:SetPoint("RIGHT", self.Frame, "RIGHT", -inset, 0);
	
	self:UpdateSize();
end

function addon:UpdateFrameFonts()
	local font = lsm:Fetch("font", db.Frame.Text.Font);
	local size = db.Frame.Text.Size;
	local flags = db.Frame.Text.Outline;
	local color = db.Frame.Text.Color;
	for i, v in ipairs(self.TextFrames) do
		v.Text:SetFont(font, size, flags);
		v.Text:SetTextColor(color.r, color.g, color.b, color.a);
	end
end

function _TableDump(arr, level)
	local text, value, padding, item = "", nil, "", nil;
	if level == nil then
		level = 0;
	end
	if level > 0 then
		for i = 0, level do
			padding = padding .. "    ";
		end
	end
	if type(arr) == "table" then
		for item, value in pairs(arr) do
			if type(item) == 'string' then
				item = '"' .. item .. '"';
			elseif type(item) == 'number' then
				item = tostring(item);
			end
			if type(value) == 'table' then
				text = text .. padding .. '[' .. tostring(item) .. '] = {\n' .. _TableDump(value, level + 1) .. padding .. "},\n";
			elseif type(value) == 'string' then
				text = text .. padding .. '[' .. tostring(item) .. '] = "' .. value .. '",\n';
			else
				text = text .. padding .. '[' .. tostring(item) .. '] = ' .. (value and tostring(value) or 'nil') .. ',\n';
			end
		end
	end
	return text;
end

function addon:TableDump(arr)
	local str = _TableDump(arr, 0);
	for v in str:gmatch("[^\n]+") do
		print(v);
	end
end